import json
import os
import sys

import boto3
import botocore
import click

from max_ard.ard_collection import ARDCollection
from max_ard.exceptions import BucketInitError, BucketRevokeError
from max_ard.io import QlrDoc
from max_ard.storage import init_bucket, revoke_bucket


@click.group()
def storage():
    """Commands related to ARD data storage"""
    pass


@click.command()
@click.argument("location")
@click.option("--profile", help="Use this AWS profile for access")
def init(location, profile=None):
    """Creates an S3 bucket if LOCATION does not exist, then authorizes the ARD data writer by adding a PutObject statement to the bucket policy of LOCATION"""
    try:
        init_bucket(location, profile=profile)
    except Exception as e:
        click.secho(f"There was a problem initializing the bucket: {str(e)}", err=True, fg="red")


@click.command()
@click.argument("location")
@click.option("--profile", help="Use this AWS profile for access")
def revoke(location, profile=None):
    """Removes the bucket policy statement from LOCATION, preventing the ARD writer from writing to the bucket"""
    try:
        revoke_bucket(location, profile=profile)
    except Exception as e:
        click.secho(
            f"There was a problem revoking access to the bucket: {str(e)}", err=True, fg="red"
        )


@click.command()
@click.argument("location")
@click.option("--dest", help='Output filename, "<filename>.qlr" for QGIS Layer Definition')
@click.option("--format", help='Format to send to stdout, "qlr" for QGIS Layer Definition')
@click.option("--aoi", help="WKT geometry to filter tiles to")
@click.option(
    "--acq_id",
    "acq_id_in",
    multiple=True,
    help="Limit result to this Acquisition ID (can use more than once",
)
@click.option(
    "--zone",
    "zone_in",
    multiple=True,
    help="Limit tiles to those in this zone. Can be used multiple times. This option can speed up scan time of large collections.",
)
@click.option("--earliest_date", help="Return results from on or after this date (YYYY-MM-DD)")
@click.option("--latest_date", help="Return results from this date or earlier (YYYY-MM-DD")
@click.option("--profile", help="Use this AWS profile for access")
@click.option(
    "--public", "anon", is_flag=True, help="Access a public bucket without using credentials"
)
def describe(**kwargs):
    """Gives a summary of ARD tiles in LOCATION (S3 or local path)"""

    format = kwargs.pop("format", None)
    dest = kwargs.pop("dest", None)
    location = kwargs.pop("location", None)
    c = ARDCollection(location, **kwargs)

    # pluralizer
    def s(i):
        s = "s"
        try:
            i = len(i)
        except:
            pass
        if i == 1:
            s = ""
        return s

    if format is None and dest is None:

        click.secho(f"Examining ARD data in {location}...", fg="cyan")
        # this will trigger access, see if it fails
        try:
            count = len(c.acq_ids)
        except PermissionError:
            click.secho(
                f"Your credentials for accessing {location} are invalid or expired",
                err=True,
                fg="red",
            )
            raise click.Abort()
        if count == 0:
            click.secho("No ARD tiles found", fg="red")
            raise click.Abort()
        click.secho(f"{count} acquisition{s(count)}", fg="cyan")
        click.secho(f"{len(c.tiles)} tile{s(c.tiles)}", fg="cyan")
        click.secho(f"Earliest date: {c.start_date}", fg="cyan")
        click.secho(f"Latest date: {c.end_date}", fg="cyan")
        zones = ",".join([str(z) for z in c.zones])
        click.secho(f"UTM zone{s(c.zones)}: {zones}", fg="cyan")
    elif format == "qlr" and dest is None:
        click.secho(QlrDoc(c, kwargs["anon"]), fg="cyan")
    elif dest.lower().endswith(".qlr"):
        with open(dest, "w") as out:
            out.write(QlrDoc(c, kwargs["anon"]))
    else:
        click.secho(
            "QGIS Layer (--format qlr or --dest filename.qlr) is the only current export type",
            fg="red",
        )


storage.add_command(init)
storage.add_command(revoke)
storage.add_command(describe)



import json
import logging
import os
from json.decoder import JSONDecodeError

from shapely import wkt
from shapely.errors import WKTReadingError
from shapely.geometry import GeometryCollection, shape
from shapely.geometry.base import BaseGeometry

from max_ard.dependency_support import HAS_FIONA
from max_ard.exceptions import MissingDependency, UnknownFileType

from maxar_ard_grid import Cell

if HAS_FIONA:
    import fiona
    from fiona.collection import Collection as FionaCollection


def convert_to_shapely(input):
    """Convert input geometry or geometry filepath to Shapely shape object

    Args:
        input: a geometry object with one or more features, or filepath to such a geometry object, as:
            - a dict that follows the Python Geometry Interface (GeoJSON-like)
            - a GeoJSON feature-like dict that has a 'geometry' key
            - a Fiona dataset reader
            - a Shapely shape *or iterable of Shapely shapes* (in which case the input will be returned as is)
            - strings in WKT, GeoJSON, or GeoJSON geometry
            - maxar_ard_grid Cells

    Returns:
        Shapely shape object of input (or iterable of Shapely shapes, if that was the input)

    * Note: if the input has multiple features, the output will be a GeometryCollection *
    """

    # ignore Nones, empty strings, other falsey things

    if not input:
        return input

    # check if we already have a Shapely shape
    if isinstance(input, BaseGeometry):
        return input

    # for Cells we need to return the WGS84 representation
    # since they are natively in UTM
    if isinstance(input, Cell):
        return input.geom_WGS84

    # or an iterable of Shapely shapes
    try:
        if all([isinstance(item, BaseGeometry) for item in input]):
            return input
    except TypeError:
        pass

    # GeoJSON-like (__geo_interface__) -- single feature
    try:
        return shape(input)
    except:
        pass

    try:
        return shape(input["geometry"])
    except:
        pass

    # GeoJSON feature collection
    try:
        return GeometryCollection([shape(g["geometry"]) for g in input["features"]])
    except:
        pass

    def fiona_converter(reader):
        geom = list(reader)
        if len(geom) > 1:
            return GeometryCollection([shape(g["geometry"]) for g in geom])
        else:
            return shape(geom[0]["geometry"])

    if HAS_FIONA and type(input) == FionaCollection:
        try:
            return fiona_converter(input)
        except:
            pass

    if type(input) is not str:
        raise TypeError("Invalid object - check list of acceptable types.")

    def quiet_wkt_loads(input):
        logger = logging.getLogger("shapely.geos")
        old_level = logger.level
        logger.setLevel(logging.CRITICAL)
        try:
            geom = wkt.loads(input)
        except Exception as e:
            raise (e)
        finally:
            logger.setLevel(old_level)
        return geom

    # strings could be wkt, json, or a path to a vector file
    if os.path.exists(input):
        try:
            return quiet_wkt_loads(open(input, "r").read())
        # try to read filepath with Fiona
        except (WKTReadingError, UnicodeDecodeError):
            if not HAS_FIONA:
                raise MissingDependency("Fiona is needed to read vector formats")
            try:
                reader = fiona.open(input)
                return fiona_converter(reader)
            except:
                raise UnknownFileType(
                    "Could not open the contents of the given file, is it a valid file?"
                )

    else:

        # wkt string
        try:
            return quiet_wkt_loads(input)
        except WKTReadingError:
            pass

        # json string, convert to dict and start over
        try:
            return convert_to_shapely(json.loads(input))
        except JSONDecodeError:
            pass

        raise TypeError("String input does not match WKT or JSON, or a valid file path")



import pprint
import re
import sys

import click

from max_ard.ard_collection import ARDCollection
from max_ard.exceptions import ARDServerException, BadARDRequest
from max_ard.io import QlrDoc
from max_ard.order import Order
from max_ard.select import Select


@click.group()
def order():
    """Tools for ordering ARD products"""
    pass


@click.command()
@click.argument("order_id", required=True)
@click.option(
    "--format", help="Output the order status in other formats: `raw` for the full response JSON"
)
def status(order_id, format):
    """Gets the status of an order given the ORDER_ID"""

    try:
        order = Order.from_id(order_id)

    except Exception as e:
        click.secho(str(e), err=True, fg="red")
        sys.exit()
    if format == "raw":
        click.secho(str(order.response.dict()), fg="cyan")
    else:
        click.secho(order.state, fg="green" if order.state == "SUCCEEDED" else "cyan")


@click.command()
@click.argument("order_id", required=True)
@click.option("--format", help='Output the order in other formats: "qlr" for QGIS layer')
@click.option("--verbose", "-v", is_flag=True, help="Return more information about the Order")
def describe(order_id, format, verbose=False):
    """Outputs order details"""
    try:
        order = Order.from_id(order_id)
    except Exception as e:
        click.secho(str(e), err=True, fg="red")
        sys.exit()
    pp = pprint.PrettyPrinter(indent=4)
    if format == "qlr":
        c = ARDCollection.from_order(order)
        click.echo(QlrDoc(c))
        exit()
    if not order.finished:
        click.secho("Order has not finished running", fg="cyan")
        if verbose:
            click.secho("*******************", fg="cyan")
            click.secho("Request details:", fg="cyan")
            click.secho(pp.pformat(order.response.order.dict()), fg="cyan")
        exit()
    c = ARDCollection.from_order(order)
    click.secho(f"getting Order ID {order_id}", fg="green")
    click.secho("*******************", fg="green")
    click.secho("Order details:", fg="green")
    click.secho(pp.pformat(order.response.order.dict()), fg="cyan")
    click.secho("*******************", fg="green")
    if format is None:

        def S(i):
            s = "s"
            try:
                i = len(i)
            except:
                pass
            if i == 1:
                s = ""
            return s

        try:
            count = len(c.acq_ids)
        except PermissionError:
            click.secho(
                f"Your credentials for accessing the order are invalid or expired",
                err=True,
                fg="red",
            )
            exit()
        if count == 0:
            click.secho("No ARD tiles found", fg="red")
            exit()
        click.secho(f"{count} acquisition{S(count)}", fg="green")
        click.secho(f"{len(c.tiles)} tile{S(c.tiles)}", fg="green")
        click.secho(f"Earliest date: {c.start_date}", fg="green")
        click.secho(f"Latest date: {c.end_date}", fg="green")
        zones = ",".join([str(z) for z in c.zones])
        click.secho(f"UTM zone{S(c.zones)}: {zones}", fg="green")
    else:
        click.secho("QGIS Layer (--format qlr) is the only available format", fg="red")
    click.secho("\n\n", fg="cyan")


class NumericType(click.ParamType):
    name = "numeric"

    def convert(self, value, param, ctx):
        try:
            return float(value)
        except TypeError:
            self.fail(
                "expected string for int() conversion, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(f"{value!r} is not a value that can be converted to a float", param, ctx)


NUM_TYPE = NumericType()


@click.command()
@click.option(
    "--destination",
    required=True,
    help="Destination of the tiles: an S3 bucket and optional prefix",
)
@click.option(
    "--select-id",
    help="Order the results of Select with the Select ID, pass LISTEN to find the ID from the piped output of a Select",
)
@click.option("--select_id", help="deprecated to only use dashes in options")
@click.option(
    "--acq-id",
    "acquisitions",
    multiple=True,
    help="Order a given acquisition by acquisition ID, can be provided multiple times."
    + "If provided with a Select ID will only generate tiles for the provided IDs.",
)
@click.option(
    "--acquisition-id", "long_acquisitions", multiple=True, help="deprecated, use --acq_id"
)
@click.option(
    "--acquisition_id",
    "old_acquisitions",
    multiple=True,
    help="deprecated to only use dashes in options",
)
@click.option(
    "--intersects",
    help="Only generate tiles that intersect this geometry in WKT format, or geometry from this file path.",
)
@click.option(
    "--role-arn", help="A trusted Role ARN for the writer to assume so it can write tiles."
)
@click.option("--role_arn", help="deprecated to only use dashes in options")
@click.option(
    "--dry-run",
    is_flag=True,
    help="Submit the order for basic validation without processing the order",
)
@click.option(
    "--bbox",
    nargs=4,
    type=NUM_TYPE,
    help="Like --intersects, but takes a bounding box in the form --bbox XMIN YMIN XMAX YMAX",
)
@click.option(
    "--bba",
    type=bool,
    default=False,
    help="Use Bundle Block Adjustment to align images within this order. See docs for limitations and best practices for using BBA.",
)
@click.option(
    "--add-email",
    "emails",
    multiple=True,
    help="Add an email address to receive order notifications. Can be used multiple times to add more than one email address.",
)
@click.option(
    "--add-sns",
    "topics",
    multiple=True,
    help="Add an AWS SNS topic to send nofications to. Can be used multiple times to add more than one topic.",
)
@click.option("--verbose", "-v", is_flag=True, help="Return more information about the Order")
@click.option("--yes", "-y", is_flag=True, help="Skip order confirmation")
def submit(**kwargs):
    """Submit an ARD Order"""

    # options that don't get sent to the API
    emails = kwargs.pop("emails", [])
    topics = kwargs.pop("topics", [])
    verbose = kwargs.pop("verbose", False)
    confirm_order = not kwargs.pop("yes", False)

    # TODO: clean this up after old args are removed
    # note the API wants the underscore so we will need to map it to the underscored name
    # will do this in Click but for now being explicit is less confusing

    # deprecations warnings to the user
    if kwargs.get("select_id", None) is not None:
        click.secho(
            "Note: `--select_id` will be deprecated, use `--select-id` instead", fg="yellow"
        )

    if kwargs.get("role_arn", None) is not None:
        click.secho(
            "Note: `--role_arn` will be deprecated, use `--role-arn` instead.", fg="yellow"
        )

    if kwargs.get("acquisition_id", None) is not None:
        click.secho(
            "Note: `--acquisition_id` will be deprecated, use `--acq-id` instead.", fg="yellow"
        )

    if kwargs.get("acquisition-id", None) is not None:
        click.secho(
            "Note: `--acquisition-id` will be deprecated, use `--acq-id` instead.", fg="yellow"
        )

    # move select-id to select_id
    select_arg = kwargs.pop("select-id", None)
    if select_arg is not None:
        kwargs["select_id"] = select_arg

    # move role-arn to role_arn
    role_arg = kwargs.pop("role-arn", None)
    if role_arg is not None:
        kwargs["role_arn"] = role_arg

    # gather up any acquisitions
    acquisitions = set()

    for arg in ["acquisitions", "long_acquisitions", "old_acquisitions"]:
        acqs = kwargs.pop(arg, None)
        if acqs is not None:
            for a in acqs:
                acquisitions.add(a)

    kwargs["acquisitions"] = list(acquisitions)

    # Listen for a select ID on stdin

    if kwargs["select_id"] is not None and kwargs["select_id"].lower() == "listen":
        click.secho(f"Listening...", fg="cyan")
        stdin = click.get_text_stream("stdin")
        stderr = click.get_text_stream("stderr")
        if stderr.readable():
            err_msg = stderr.read()
            if len(err_msg.strip()) > 0:
                click.secho(f"Error on stderr: {err_msg}", err=True)
                exit()
        try:
            select_id = re.match(r"Select (\d+)", stdin.read())[1]
            click.secho(f"Input stream matched select {select_id}", fg="cyan")
        except TypeError:
            click.secho(
                f"Input stream was not usable, did the Select succeed?", fg="red", err=True
            )
            exit()
        try:
            select = Select.from_id(select_id)
        except ARDServerException as e:
            msg = "A Select ID was found, but there was an error fetching its details: \n"
            msg += str(e)
            click.secho(msg, fg="red", err=True)
            exit()
        select.wait_for_success()
        kwargs["select_id"] = select_id
        click.secho(f"Ordering from Select {select_id}", fg="cyan")
    try:
        order = Order(**kwargs)
    except Exception as e:
        click.secho(f"There was an error in your parameters: {e}", fg="red", err=True)
        exit()

    if not emails and not topics:
        if confirm_order:
            if input("No notifications supplied, do you still want to order?") != "y":
                exit()
    for email in emails:
        order.add_email_notification(email)
    for topic in topics:
        order.add_sns_notification(topic)

    if verbose:
        pp = pprint.PrettyPrinter(indent=4)
        click.secho(f"Request payload:", fg="cyan")
        click.secho(pp.pformat(order.request.to_payload()), fg="cyan")

    if kwargs["dry_run"]:
        click.secho("Checking order in dry-run mode...", fg="cyan")
        try:
            order.submit()
        except BadARDRequest as e:
            click.secho(f"There was an error in your request: {e}", err=True, fg="red")
            exit()
        click.secho("Dry run successful", fg="green")
    else:
        if confirm_order:
            if input("Are you sure you want to order? (y/n) ") != "y":
                exit()
        try:
            order.submit()
            click.secho(f"Order {order.order_id} has been submitted", fg="cyan")
            click.secho(
                f"Run `max-ard order status {order.order_id}` to check the status", fg="cyan"
            )
            click.secho('When the status is "SUCCEEDED" you can: ', fg="cyan")
            click.secho(
                f"Run `max-ard order describe {order.order_id}` to see a basic overview of the results",
                fg="cyan",
            )
            click.secho("Run `max-ard order` to see all the `order` commands", fg="cyan")
        except BadARDRequest as e:
            click.secho(f"There was an error in your request: {e}", err=True, fg="red")


order.add_command(status)
order.add_command(describe)
order.add_command(submit)



"""Helpers for optional dependencies and platform quirks
"""
import os
import platform

# Conda on Windows proj data
if platform.system() == "Windows":
    # turn on CDN backup files
    os.environ["PROJ_NETWORK"] = "ON"
    # set to Conda's share location
    conda_prefix = os.environ.get("CONDA_PREFIX")
    if conda_prefix:
        os.environ["PROJ_LIB"] = os.path.join(conda_prefix, "Library", "share", "proj")


# Optional dependencies
try:
    import rasterio

    HAS_RASTERIO = True
except ImportError:
    HAS_RASTERIO = False

try:
    import fiona

    HAS_FIONA = True
except ImportError:
    HAS_FIONA = False



import json

import click
from maxar_ard_grid import Cell, covers


@click.group()
def cell():
    """Subcommand for ARD Grid cell tools"""

    pass


@click.command()
@click.argument("quadkey", required=True)
@click.argument("zone", required=False)
@click.option("--format", help="Output the cell in different formats: `geojson` for GeoJSON")
def describe(quadkey, zone, format):
    """Output a description of a cell given a QUADKEY and optionally a ZONE

    QUADKEY by itself must be a full Cell ID: z15-230232110
    or
    QUADKEY is just the quadkey: 230232110 plus a ZONE : 15"""

    if zone is None:
        cell = Cell(quadkey)
    else:
        cell = Cell(quadkey, zone=zone)
    # error handling
    if format is None:
        click.secho(f"Cell {cell.quadkey} in zone {cell.zone}", fg="cyan")
        xmin, ymin, xmax, ymax = cell.bounds
        click.secho("UTM bounds: ", fg="cyan")
        click.secho(f" Lower left corner: {xmin}, {ymin}", fg="cyan")
        click.secho(f" Upper right corner: {xmax}, {ymax}", fg="cyan")
        click.secho("Geographic bounds:", fg="cyan")
        xmin, ymin, xmax, ymax = cell.geom_WGS84.bounds
        click.secho(f" Lower left corner: {xmin}, {ymin}", fg="cyan")
        click.secho(f" Upper right corner: {xmax}, {ymax}", fg="cyan")

        click.secho(f"Cell size: {xmax - xmin}m", fg="cyan")
    if format == "geojson":
        click.secho(cell.to_GeoJSON(), fg="cyan")


@click.command(name="covers")
@click.argument("geom", type=str)
@click.option("--zoom", type=int, help="Zoom of cells to calculate, default is 12")
@click.option("--format", help="Output the cellis in different formats: `geojson` for GeoJSON")
def _covers(geom, format, zoom):
    """Calculate the cell IDs that cover a given geometry GEOM. GEOM can be WKT or GeoJSON"""

    if zoom is None:
        cells = covers(geom)
    else:
        cells = covers(geom, zoom=zoom)
    if format is None:
        for cell in cells:
            click.secho(cell.id, fg="cyan")
    if format == "geojson":
        fc = {"type": "FeatureCollection", "features": []}
        for cell in cells:
            fc["features"].append(cell.to_feature())
        click.secho(json.dumps(fc), fg="cyan")


cell.add_command(_covers)
cell.add_command(describe)



""" ARD Monitors

ARD Monitors watch AOIs and send notifications when new imagery is acquired. Monitors 
can be created with queries to restrict which acquisitions trigger a notification.

Provides
--------

- Create, retrieve, list, and delete monitors

Notes
-----

Monitors currently support the following notification types:
- Email
- Amazon SNS

Examples
-------

Create a monitor that will watch a bounding box for new imagery collected
by Worldview 1 and send an email when new acquisitions are available:

>>> from max_ard import Monitor

>>> bbox = [-120.1, 43.6, -120.7, 44.2]

>>> query = query = {"platform": {"eq": "worldview-01"}}

>>> monitor = Monitor(bbox=bbox, query=query)

>>> monitor.add_email_notification("me@example.com")

>>> monitor.submit()
"""

import warnings
from typing import List, Union, Any

from maxar_ard_grid import Cell
from max_ard.base_collections import (
    ARDModel,
    Submitted,
    hydrate_with_responses,
    EmailNotification,
    SNSNotification,
)
from max_ard.exceptions import MissingARDResource, NotSubmitted
from max_ard.session import ard_url, get_user_session, link_paginated_response
from max_ard.io import convert_to_shapely
from pydantic import BaseModel
from shapely.geometry import mapping, MultiPolygon


class MonitorAction(BaseModel):
    """Model for monitoring actions

    Note: The only action so far is "notify"
    """

    action: str
    notifications: List[Union[EmailNotification, SNSNotification]] = []


class MonitorRequest(ARDModel):
    """Model of Monitor API Request to send"""

    name: str = None
    description: str = None
    # filter
    intersects: str = None
    bbox: List = None
    # destination
    query: dict = None
    actions: List[MonitorAction] = []

    class Config:
        extra = "allow"
        validate_assignment = True


class MonitorResponse(BaseModel):
    """Model of Monitor API Response"""

    cell_ids: List[str]
    monitor_id: str
    account_id: str
    user_id: str
    name: str = None
    description: str = None
    active: bool
    intersects: Union[str, dict] = None
    bbox: List = None
    query: dict = {}
    actions: List[dict]
    created: str

    class Config:
        extra = "allow"
        validate_assignment = True


class Monitor(Submitted):
    """Object for creating and retrieving ARD Monitors

    Parameters
    ----------
    intersects : Geometry-like objects, str (optional))
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north]
    name : str
        Optional name for the monitor (display only)
    description : str
        Optional description for the monitor (display only)
    query : dict
        Query dictionary
    session : session object, optional
        A user_session or application_session object


    Notes
    -----

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading
    """

    def __init__(
        self, intersects=None, bbox=None, name=None, description=None, query={}, session=None
    ):

        self.session = session or get_user_session()
        if intersects is not None:
            intersects = convert_to_shapely(intersects).wkt
        self.request = MonitorRequest(
            name=name, description=description, intersects=intersects, bbox=bbox, query=query
        )
        self.response = None

    def add_email_notification(self, address: str) -> None:
        """Add an email notification to the Monitor

        Parameters
        ----------
        address : str
            Email address to receive order notifications

        Returns
        -------
        None
        """

        notification = EmailNotification(address=address)
        self.request.actions.append(MonitorAction(action="notify", notifications=[notification]))

    def add_sns_notification(self, topic_arn: str) -> None:
        """Add an AWS SNS notification topic to receive Monitor notifications

        Parameters
        ----------
        topic_arn : str
            AWS SNS topic ARN to recieve order notifications

        Returns
        -------
        None
        """

        notification = SNSNotification(topic_arn=topic_arn)
        self.request.actions.append(MonitorAction(action="notify", notifications=[notification]))

    @property
    def submitted(self) -> bool:
        """Submittal state of the monitor"""
        if self.response is None:
            return False
        else:
            return self.response.monitor_id is not None

    @Submitted.required
    def get_history(self) -> List[dict]:
        """Get the history of image matches for this monitor

        Parameters
        ----------
        None

        Returns
        -------
        list
            List of match data dictionaries

        Notes
        -----
        This is not intended to be polled for matches - you can use Select or the Metadata API to check
        for new results.
        """

        url = ard_url("monitor", "history", self.monitor_id)
        id_key = lambda monitor: monitor["monitor_id"]
        monitors = link_paginated_response(self.session, url, id_key)

        return monitors  # ?

    @classmethod
    def list_monitors(
        cls, session=None, raw: bool = False, **kwargs
    ) -> List[Union[dict, "Monitor"]]:
        """List monitors

        Parameters
        ----------
        session: Session
            A user_session or application_session objectj, default uses current user
        raw: bool
            Returns dictionaries if True, Default is to return Monitor objects

        Keywords
        --------
        kwargs
            Pass-through HTTP parameters

        active_only: bool, default is True
            Only show active monitors

        Returns
        -------
        list of dict or Monitor
            Monitors
        """

        url = ard_url("monitor")
        session = session or get_user_session()
        try:
            responses = link_paginated_response(session, url, **kwargs)
        except MissingARDResource:
            return []

        if raw:
            return responses
        else:
            return hydrate_with_responses(Monitor, MonitorResponse, responses, session=session)

    @property
    @Submitted.required
    def monitor_id(self) -> str:
        """Monitor ID"""
        return self.response.monitor_id

    @property
    @Submitted.required
    def cells(self) -> List[Cell]:
        """List of cells being watched by the monitor, as maxar_ard_grid Cell objects"""
        return [Cell(cell) for cell in self.response.cell_ids]

    @property
    def notifications(self) -> List[Union[EmailNotification, SNSNotification]]:
        "List of notifications for this monitor"
        return [action["notification"] for action in self.actions if action["action"] == "notify"]

    def __getattr__(self, name: str) -> Any:
        if self.submitted:
            return getattr(self.response, name)
        try:
            return getattr(self.request, name)
        except AttributeError as e:
            if name not in MonitorResponse.__fields__.keys():
                raise e

        raise NotSubmitted(f"Can't access `{name}` until Monitor has been submitted")

    @classmethod
    def from_id(cls, monitor_id: str) -> "Monitor":
        """Retrieve a Monitor by its ID

        Parameters
        ----------
        monitor_id: str
            Monitor ID

        Returns
        -------
        Monitor
            Requested Monitor

        Raises
        ------
        MissingARDResource
            The ID can not be found.
        """
        response_obj = cls.get_monitor(monitor_id)["monitor"]
        return cls.from_response(response_obj)

    @classmethod
    def from_response(cls, response: dict) -> "Monitor":
        """Create a Monitor object using an API response dictionary of attributes

        Parameters
        ----------
        response: dict
            Monitor attribute dict

        Returns
        -------
        Monitor

        """
        instance = cls()
        instance.response = MonitorResponse(**response)
        return instance

    @staticmethod
    def get_monitor(monitor_id: str, session=None) -> dict:
        """Retrieve a Monitor API payload by its ID

        Parameters
        ----------
        monitor_id: str
            Monitor ID
        session: Session
            A user_session or application_session objectj, default uses current user

        Returns
        -------
        dict
            Requested monitor API payload

        Raises
        ------
        MissingARDResource
            The ID can not be found.
        """
        session = session or get_user_session()
        url = ard_url("monitor", "config", monitor_id)
        return session.get(url).json()

    @staticmethod
    def send_monitor(payload: dict, session=None) -> dict:
        """Post a Monitor API payload

        Parameters
        ----------
        payload: dict
            API monitor payload
        session: Session
            A user_session or application_session objectj, default uses current user

        Returns
        -------
        dict
            Created monitor API payload
        """

        session = session or get_user_session()
        url = ard_url("monitor")

        return session.post(url, json=payload).json()

    def submit(self) -> None:
        """Submit this monitor to the API"""

        if self.submitted:
            warnings.warn("This select has already been submitted")
            return

        payload = self.request.to_payload()
        response = self.send_monitor(payload, session=self.session)
        self.response = MonitorResponse(**response["monitor"])

    @Submitted.required
    def delete(self) -> None:
        """Soft delete this monitor from the API.

        This sets the monitor's `active` flag to `False`. Inactive monitors are not returned
        by the list endpoint unless the parameter `active_only=False` is sent.

        An inactive monitor does not run but still can be fetched by its ID.

        """
        url = ard_url("monitor", "config", self.monitor_id)
        self.session.delete(url)

    def __repr__(self) -> str:
        if self.submitted:
            monitor_id = self.monitor_id
        else:
            monitor_id = "(unsubmitted)"
        if self.name is None:
            name = ""
        else:
            name = f": '{self.name}'"
        return f"<Monitor {monitor_id}{name}>"

    @property
    def __geo_interface__(self):
        """Python Geospatial Interface"""
        geoms = [cell.geom_WGS84 for cell in self.cells]
        return mapping(MultiPolygon(geoms))



"""Convert ARD Selects to a KMZ File"""

import warnings

warnings.warn(
    "max_ard.outputs.kmz_format.KmzDoc has been moved to max_ard.io.KmzDoc",
    DeprecationWarning,
    stacklevel=2,
)

from max_ard.io import KmzDoc as NewKmzDoc


def KmzDoc(select, path):
    """Convert a Select to a KMZ file and save

    Parameters
    ----------
    select: Select
    path: path at which to save the KMZ

    Returns
    -------
    none"""

    warnings.warn(
        "max_ard.outputs.kmz_format.KmzDoc has been moved to max_ard.io.KmzDoc",
        DeprecationWarning,
        stacklevel=2,
    )
    return NewKmzDoc(select, path)



"""Export ARD collections to other file formats

Provides
--------
- QGIS Layer export of ARDCollections
- KML, KMZ, SHP exports of Selects

Notes
-----
Selects and ARDCollections export to different formats due to the underlying data formats.

There is not currently a way for Google Earth to read the ARD COGs directly. For KMLs to display 
would require either thumbnails to be produced of the COGs, or some sort of KML SuperOverlay
web service to be set up.

It might be possible to display selects in QGIS, but since the Select browse image preview tiles
do not have georeferencing files it would take some experimentation to correctly configure
their layers, or add the ability to the preview endpoint to answer requests for georeferencing files
"""

from .inputs import convert_to_shapely
from .kml_format import KmlDoc
from .kmz_format import KmzDoc
from .qlr_format import QlrDoc
from .shp_format import ShpDoc



""" ARD Exceptions """

from datetime import datetime

#########
# General Errors
#########


class MissingDependency(Exception):
    """A dependency required for this functionality has not been installed"""

    pass


#########
# API Errors
#########


class ARDServerException(Exception):
    """The ARD API service encountered an error"""

    def __init__(self, msg, *args, **kwargs):
        msg = f"{msg} ({datetime.now()})"
        super().__init__(msg, *args, **kwargs)


class BadARDRequest(Exception):
    """The ARD API request sent was incorrectly formatted"""

    pass


class MissingARDResource(Exception):
    """An ARD API request returned a 404"""

    pass


class MissingSummary(MissingARDResource):
    """Summary data for a requested cell does not exist.
    The cell needs to be registered with the metadata service."""

    pass


class GeometryException(Exception):
    """The ARD API request sent was incorrectly formatted"""

    pass


class OversizeRequestException(Exception):
    """The ARD API request sent was incorrectly formatted"""

    pass


class UnAuthorizedException(Exception):
    """The ARD API request sent had bad creds"""

    pass


class NotSubmitted(Exception):
    """A Select or Order has not been submitted"""

    pass


class NotFinished(Exception):
    """A select has not finished running so results are not available yet"""

    pass


class SelectError(Exception):
    """Errors returned from the Select Service"""

    pass


class NotSuccessful(Exception):
    """Error in processing"""

    pass


#########
# IO Errors
#########


class UnknownFileType(Exception):
    """A file input was not recognized as a readable format"""

    pass


#######
# S3 Storage Errors
#######


class BucketInitError(Exception):
    """An error occured initializing an S3 bucket"""

    pass


class BucketRevokeError(Exception):
    """An error occured initializing an S3 bucket"""

    pass


########
# Hook functions for Request.Session.hooks
########


def oversize_request_hook(r, *args, **kwargs):
    """
    Hook for handling oversize request errors from API.
    """
    if r.status_code == 413 and "Request Entity Too Large" in r.json()["message"]:
        raise OversizeRequestException(
            f'{r.json()["message"]}, Reduce size of attributes and try again.'
        )

    return r


def bad_geom_request_hook(r, *args, **kwargs):
    """
    Hook for handling bad geometry errors from API.
    """
    if r.status_code == 400 and "Problem with intersects/bbox" in r.json()["message"]:
        raise GeometryException(r.json()["message"])

    return r


def bad_ard_request_hook(r, *args, **kwargs):
    """
    Hook for handling 400 errors from API.
    """
    if r.status_code == 400:
        raise BadARDRequest(r.json()["message"])

    return r


def unauth_request_hook(r, *args, **kwargs):
    """
    Hook for handling 401 errors from API.
    """
    if r.status_code == 401:
        raise UnAuthorizedException(
            f"Server response was {r.json()['message']}, check your credentials and try again."
        )

    return r


def missing_resource_hook(r, *args, **kwargs):
    if r.status_code == 404:
        raise MissingARDResource(r.json()["message"])

    return r


def ard_server_request_hook(r, *args, **kwargs):
    """
    Hook for handling non-20X errors from API.
    """
    if not r.ok:
        raise ARDServerException(r.json()["message"])

    return r



"""Convert ARD Selects to a KML File"""

import os

from shapely.geometry import shape
from shapely.geometry.polygon import orient

from max_ard.session import ard_url

__all__ = ("KmlDoc",)


def KmlDoc(select):
    """Convert a Select to a KML file

    Parameters
    ----------
    select: Select

    Returns
    -------
    str
        A KML document of the Select"""

    results = select.results
    folders = []
    acquisitions = results.acquisitions
    acquisitions.sort(key=lambda x: x.date)
    folders = (KmlFolder(acquisition) for acquisition in acquisitions)
    folders = "\n".join(folders)

    return f"""<?xml version="1.0" encoding="UTF-8"?>
    <kml xmlns="http://www.opengis.net/kml/2.2" xmlns:gx="http://www.google.com/kml/ext/2.2">
        <Document id="1"> 
            <Style id="red">
                <LineStyle>
                    <color>ff0000ff</color>
                    <colorMode>normal</colorMode>
                </LineStyle>
                <PolyStyle>
                    <colorMode>normal</colorMode>
                    <fill>0</fill>
                    <outline>1</outline>
                </PolyStyle>
            </Style>
            {folders}
        </Document>
    </kml>"""


def KmlFolder(acquisition):

    tiles = (TileLayer(tile) for tile in acquisition)
    tiles = "\n".join(tiles)

    return f"""
        <Folder>
        <name>{acquisition.acq_id}</name>
        {tiles}
    </Folder> """


def TileLayer(tile):

    return f"""
    {KmlPolygon(tile)}
    {KmlOverlay(tile)}"""


def KmlPolygon(tile):

    # make sure the geometry is CCW wound
    geom = orient(shape(tile))
    coords = [(*t, 0) for t in geom.exterior.coords]
    coords = " ".join([f"{c[0]},{c[1]},{c[2]}" for c in coords])

    return f"""
        <Placemark>
        <name>{tile.cell.id}</name>
        <styleUrl>#red</styleUrl>
        <Polygon>
            <outerBoundaryIs>
                <LinearRing>
                    <coordinates>{coords}</coordinates>
                </LinearRing>
            </outerBoundaryIs>
        </Polygon>
    </Placemark>"""


def KmlOverlay(tile):

    # browse isn't part of the ard api, don't use ard_url()
    url = ard_url("browse", "preview", tile.acq_id, tile.cell_id)
    overlay_geom = orient(shape(tile.cell.geom_WGS84))
    quad_pts = list(overlay_geom.exterior.coords)[:4]
    # these need to go ll, lr, ur, rl pairs
    # so we rotate until ll is at the front of the quad
    cx, cy = overlay_geom.centroid.coords[0]
    from shapely.geometry import box

    while quad_pts[0][0] >= cx or quad_pts[0][1] >= cy:
        quad_pts.append(quad_pts.pop(0))
    quad_pts = [(*t, 0) for t in quad_pts]
    coords = " ".join([f"{c[0]},{c[1]},{c[2]}" for c in quad_pts])

    return f"""
        <GroundOverlay>
        <name>{tile.acq_id}</name>
        <Icon>
            <href>{url}</href>
        </Icon>
        <gx:LatLonQuad>
            <coordinates>{coords}</coordinates>
        </gx:LatLonQuad>
    </GroundOverlay>"""



"""Tools and utilities for processing ARD images

Provides
--------
- read/write windows from ARD collections
"""

from collections import deque
from typing import Any

from maxar_ard_grid import covers
from maxar_ard_grid.grid import get_CRS, get_transform
from shapely.geometry import shape
from shapely.ops import transform
from shapely.prepared import prep

from max_ard.dependency_support import HAS_RASTERIO
from max_ard.exceptions import MissingDependency
from max_ard.io import convert_to_shapely

if HAS_RASTERIO:
    import numpy as np
    import rasterio
    from rasterio import windows


__all__ = ("read_windows", "write_windows")


class COGReader:
    """An optimized COG rasterio reader

    ARD COGs do not have any sidecar files, therefore when opening
    we can use GDAL_DISABLE_READDIR_ON_OPEN='EMPTY_DIR' to skip looking
    for sidecar files. This skips an S3 LIST operation which takes
    time and money (LISTs cost more than READs).

    We also override some GDAL read settings for more potential speed
    and cost gains

    Parameters
    ----------
    path : str
        Path of COG to open
    *args : list
        Any additional args to pass to rasterio.open()
    **kwargs : dict
        Any additional keyword arguments to pass to rasterio.open()"""

    COG_ENV = {
        "GDAL_HTTP_MERGE_CONSECUTIVE_RANGES": True,
        "VSI_CACHE": True,
        "GDAL_HTTP_MULTIPLEX": True,
        "VRT_SHARED_SOURCE": 0,
    }

    def __init__(self, path, *args, **kwargs):
        with rasterio.Env(
            GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR", CPL_VSIL_CURL_ALLOWED_EXTENSIONS=".tif"
        ):
            self.dsr = rasterio.open(path, *args, **kwargs)

    def __getattribute__(self, name: str) -> Any:
        """Proxy to the underlying rasterio Dataset Reader"""
        if name == "read":

            def func(*args, **kwargs):
                with rasterio.Env(**COGReader.COG_ENV):
                    return self.dsr.read(*args, **kwargs)

            return func
        else:
            dsr = object.__getattribute__(self, "dsr")
            if name == "dsr":
                return dsr
            else:
                return getattr(dsr, name)

    # context manager dunders can't be proxied
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.dsr.close()


def iterify(input):
    """Takes an input of any type and returns a list containing that input, if the input is not already an iterable or if it is a string.
    If the input is a non-string iterable, returns the input as is."""

    if isinstance(input, str):
        return [input]
    else:
        try:
            input[0]
            return input
        except:
            return [input]


def loop_geoms(collection, geoms, src_proj=4326):
    """Loops over a geometry source and emits a tuple of:
    - An ARDTile that overlaps a feature from the source
    - The feature as a Shapely geometry, reprojected to the tile's UTM projection

    Args:
        geoms: one or more geometry objects as:
            - Shapely geometry objects
            - dicts that follow the Python Geometry Interface (GeoJSON-like)
            - GeoJSON feature-like dicts that have a 'geometry' key
            - lists or tuples of above
            - a Fiona dataset reader
            - a path to a readable geometry file.

    Yields:
        tuple(ARDTile, shape): tuple of ARD Tile and reprojected Shapely geometry object"""

    src_crs = get_CRS(src_proj)

    geoms = convert_to_shapely(geoms)
    for geom in iterify(geoms):
        for cell in covers(geom, src_proj=src_proj):
            for tile in collection.get_stack(cell):
                tile_crs = get_CRS(tile.cell.proj)
                tfm = get_transform(src_crs, tile_crs)
                proj_geom = transform(tfm, geom)
                yield tile, proj_geom


def read_windows(collection, geoms, src_proj=4326):
    """A generator for reading windows from overlapping tiles in an ARDCollection

    Args:
        geoms: one or more geometry objects as:
            - Shapely geometry objects
            - dicts that follow the Python Geometry Interface (GeoJSON-like)
            - GeoJSON feature-like dicts that have a 'geometry' key
            - lists or tuples of above
            - a Fiona dataset reader
        src_proj: An EPSG identifier parseable by Proj4

    Yields:
        Tuple of:
            ARDTile of tile
            Shapely polygon representing the bounds of the input geometry
                in the tile's UTM coordinate system
            a Reader function:

                reader(asset)

                    Args:
                        asset(str): asset to read: `visual`, `pan`, or `ms`

                    Returns:
                        numpy array of data covered by the returned geometry

    Example:

    geoms = [...]
    for tile, geom, reader in read_windows(geoms):
        # can check tile.properties, or fetch tile.clouds
        # can check if `geom` intersects with clouds
        data = reader('visual')"""

    if not HAS_RASTERIO:
        raise MissingDependency("Rasterio is required to read ARD rasters")

    for tile, geom in loop_geoms(collection, geoms, src_proj):

        def reader(asset, **kwargs):
            dsr = tile.open_asset(asset)
            window = windows.from_bounds(*geom.bounds, dsr.transform)
            return dsr.read(window=window, **kwargs)

        yield (tile, geom, reader)


def write_windows(collection, geoms, src_proj=4326):

    if not HAS_RASTERIO:
        raise MissingDependency("Rasterio is required to read ARD rasters")

    for tile, geom in loop_geoms(collection, geoms, src_proj):

        def writer(asset, name, **kwargs):
            dsr = tile.open_asset(asset)
            window = windows.from_bounds(*geom.bounds, dsr.transform)
            profile = dsr.meta
            profile.update(
                {
                    "width": window.width,
                    "height": window.height,
                    "count": dsr.count,
                    "transform": dsr.window_transform(window),
                    "driver": "GTiff",
                }
            )
            profile.update(kwargs)
            with rasterio.open(name, "w", **profile) as dest:
                dest.write(dsr.read(window=window))

        yield (tile, geom, writer)


class AcquisitionReader:
    """A pixel reader that reads across tiles in an acquisition.

    Default behavior is to histogram match pixels from different tiles to match
    the tile that represents the majority of pixels in the window.

    NOTE: experimental and not production tested!

          Will have problems with acquisitions that span UTM zones.

          This can be probably be fixed with rasterio.warpedVRTs but
          will take more work if there's demand for this feature.

    Parameters
    ----------
    acquisition: max_ard.Acquisition
        Acquisition object"""

    def __init__(self, acquisition):
        # store the tile cell bounds as prepared geoms
        # for intersection
        if not HAS_RASTERIO:
            raise MissingDependency("Rasterio is required for pixel reading")
        self.cells = {tile: prep(shape(tile.cell)) for tile in acquisition}

    def read(self, aoi, asset, src_proj=4326, match=True):
        """Read an AOI window from an asset type

        Parameters
        ----------
        aoi: almost any geometry input
            Window to read, will use this geometry's envelope if not rectilinear.

        asset: str
            Asset name, currently 'visual', 'pan', 'ms;

        match: bool, optional
            Whether to apply histogram matching.

        Returns
        -------
        ndarrary
            Array of image data"""

        # Priority in the image goes to the source tile with the most pixels
        # Calculate overlapping areas for tiles and sort biggest to smallest
        candidates = []
        src_crs = get_CRS(src_proj)

        aoi = convert_to_shapely(aoi)

        for tile, geom in self.cells.items():
            tile_crs = get_CRS(tile.cell.proj)
            tfm = get_transform(src_crs, tile_crs)
            proj_aoi = transform(tfm, aoi)
            # if a cell contains the AOI just return it
            if geom.contains(proj_aoi):
                return self.read_aoi(tile, asset, proj_aoi)[0]
            elif geom.intersects(proj_aoi):
                candidates.append((tile, shape(tile.cell).intersection(proj_aoi).area))
        if len(candidates) == 0:
            raise ValueError("AOI does not intersect any tiles")
        # we need to grab the biggest first, then insert based on area
        candidates.sort(key=lambda x: x[1])
        candidates = deque(candidates)
        candidates.rotate(1)

        output = None
        reference = None
        # Loop through sources and insert into output array
        for tile, area in candidates:
            array, window = self.read_aoi(tile, asset, proj_aoi)
            bands, height, width = array.shape
            if output is None:
                output = np.zeros(
                    (bands, int(round(window.height)), int(round(window.width))), array.dtype
                )

            if window.col_off < 0:
                col_off = -int(round(window.col_off))
            else:
                col_off = 0
            if window.row_off < 0:
                row_off = -int(round(window.row_off))
            else:
                row_off = 0
            # first loop grabs the reference (biggest) image
            if reference is None:
                reference = array
                reference_position = (row_off, col_off, height, width)
            else:
                if match:
                    array = self.histogram_match(array, reference)
                output[:, row_off : row_off + height, col_off : col_off + width] = array

        # insert the reference back on top
        row_off, col_off, height, width = reference_position
        output[:, row_off : row_off + height, col_off : col_off + width] = reference
        return output

    def read_aoi(self, tile, asset, aoi):
        """Read an AOI from a tile

        Parameters
        ----------
        tile: max_ard.ARDTile
            Tile to read from

        asset: str
            Asset name, currently 'visual', 'pan', 'ms;

        aoi: Shapely geometry
            Window to read, will use this geometry's envelope if not rectilinear.
            Must be in UTM to match the acquisition (also see notes above)


        match: bool, optional
            Whether to apply histogram matching.

        Returns
        -------
        ndarrary
            Array of image data"""

        source = tile.open_asset(asset)
        window = windows.from_bounds(*aoi.bounds, source.transform)
        return (source.read(window=window), window)

    def histogram_match(self, source, template):
        """Adjust the pixel values of an image such that its histogram
        matches that of a target image.

        from
        https://gist.github.com/jcjohnson/e01e4fcf7b7dfa9e0dbee6c53d3120b6)

        Code adapted from
        http://stackoverflow.com/questions/32655686/histogram-matching-of-two-images-in-python-2-x

        Parameters
        ----------
        source: np.ndarray
            Image to adjust
        template: np.ndarray
            Template image; can have different dimensions to source

        Returns
        -------
        np.ndarray
            The transformed output image"""

        oldshape = source.shape
        source = source.ravel()
        template = template.ravel()

        # get the set of unique pixel values and their corresponding indices and
        # counts
        s_values, bin_idx, s_counts = np.unique(source, return_inverse=True, return_counts=True)
        t_values, t_counts = np.unique(template, return_counts=True)

        # take the cumsum of the counts and normalize by the number of pixels to
        # get the empirical cumulative distribution functions for the source and
        # template images (maps pixel value --> quantile)
        s_quantiles = np.cumsum(s_counts).astype(np.float64)
        s_quantiles /= s_quantiles[-1]
        t_quantiles = np.cumsum(t_counts).astype(np.float64)
        t_quantiles /= t_quantiles[-1]

        # interpolate linearly to find the pixel values in the template image
        # that correspond most closely to the quantiles in the source image
        interp_t_values = np.interp(s_quantiles, t_quantiles, t_values)

        return interp_t_values[bin_idx].reshape(oldshape)



"""Convert ARD Selects to a KMZ File"""

from os import mkdir, rename
from os.path import join, split, splitext
from shutil import make_archive, rmtree

from max_ard.io import KmlDoc

__all__ = ("KmzDoc",)


def KmzDoc(select, path):
    """Convert a Select to a KMZ file and save

    Parameters
    ----------
    select: Select
    path: path at which to save the KMZ

    Returns
    -------
    none"""

    # get the kml doc as a string
    kml_str = KmlDoc(select)

    splitext_path = splitext(path)
    path_minus_extension = splitext_path[0]

    split_path = split(path)
    basepath = split_path[0]
    kmz_fullname = split_path[1]

    kmz_basename = splitext(kmz_fullname)[0]
    folder_path = join(basepath, kmz_basename)

    mkdir(folder_path)

    kml_path = join(folder_path, kmz_basename + ".kml")
    with open(kml_path, "w") as out:
        out.write(kml_str)

    make_archive(path_minus_extension, "zip", folder_path)
    folder_zipped_path = folder_path + ".zip"

    rename(folder_zipped_path, folder_path + ".kmz")
    rmtree(folder_path)



"""Search ARD tile metadata

Provides
--------
  1. A MetaSelect class for Select-like searches performed at the metadata level
  2. Functions for tile metadata searches
  3. An implementation of a multithreaded batch search function
  4. Functions for tile summary metadata searches.


Documentation
-------------

Coming soon

Examples
-------

Using a search function that returns GeoJSON FeatureCollection formatted data

    >>> from max_ard.metadata import search_by_area, MetaSelect
    >>> bbox = [-106.8, 35.1, -106.4, 35.4]
    >>> datetime =  "2020-07-01T00:00:00Z/2021-01-25T00:00:00Z"
    >>> query = {
            "platform": {
            "eq": "worldview-02"
            },
            "aoi:cloud_free_percentage": {
            "gte": 95
            },
            "aoi:data_percentage": {
            "gte": 75
            }
        }

    >>> tiles = search_by_area(datetime=datetime, bbox=bbox, query=query, stack_depth=3)
    >>> tiles['features'][0]

Using a MetaSelect for a similar interface to Select objects

    >>> select = MetaSelect(datetime=datetime, bbox=bbox, query=query, stack_depth=3)
    >>> results = s.results
    >>> print(results)
    < SelectResult (95 tiles in 12 acquisitions) >
    >>> results.dates
    ['2020-07-10',
    '2020-07-24',
    '2020-07-29',
    '2020-09-08',
    '2020-10-21',
    '2021-01-16']
 
Get summary metadata for a tile

    >>> summarize_cell(id, id)

"""


from concurrent.futures import as_completed
from copy import deepcopy
from typing import List, Union, Any

from requests_futures.sessions import FuturesSession
from shapely.geometry import box
from shapely.ops import unary_union
from shapely.prepared import prep

from max_ard.io import convert_to_shapely
from max_ard.select import SelectResult
from max_ard.session import get_user_session, ard_url
from max_ard.exceptions import (
    ARDServerException,
    MissingSummary,
    MissingARDResource,
    GeometryException,
)
from maxar_ard_grid import covers


###
# MetaSelect Objects
###


class MetaSelect:
    """An ARD API MetaData Select object

    Parameters
    ----------
    acq_ids : iterable of str
        An iterable of acquisition IDs to search for
    datetime : str
        Date or date range string
    intersects : Geometry-like objects, str (optional))
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north]
    query : dict
        Query dictionary
    stack_depth: int or None, optional
        Maximum number of tiles to return, default is all tiles
    image_age_category : iterable of str, optional
        One or
    session : session object, optional
        A user_session or application_session object

    Attributes
    ----------

    results: Results
        A container object holding ARDtile objects

    Notes
    -----

    The `MetaSelect` object works like a standard `Select` object but uses the SDK's batch select functionality to retrieve results faster. Since `MetaSelects` break up the search work and do not generate output files, they generally will generate results over large areas that can cause a regular Select to fail.

    There are a few differences in how `MetaSelects` work:

    - Default stack depth is unlimited, use `stack_depth` to set a limit on tiles returned.
    - There is no scoring on query parameters. Instead tiles are returned sorted newest to oldest. A `stack_depth` of 5 will return the five most recent matching tiles. If `stack_depth` is not provided, all results will be returned.
    - The search is run when the object is initialized. You do not need to call `submit()` or `wait_for_success()`, however for compatibility both methods can be called without raising errors. The same applies to properties like `succeeded` - they are set to the appropriate values for compatibility but if the object successfully initializes, it has already completed the search process.
    - Usage data is not calculated.
    - No file artifacts are generated, so the following methods and attributes will raise NotImplemented errors:

    - select_id
    - from_id()
    - get_select()
    - usage
    - get_signed_link()
    - get_link_contents()
    - copy_link()

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading"""

    def __init__(
        self,
        acq_ids: List[str] = None,
        datetime: str = None,
        intersects=None,
        bbox: list = None,
        query: dict = {},
        stack_depth: int = None,
        image_age_category: List[str] = None,
        session=None,
    ):

        kwargs = {
            "filters": [],
            "session": session,
            "acq_ids": acq_ids,
            "datetime": datetime,
            "intersects": intersects,
            "bbox": bbox,
            "query": query,
            "stack_depth": stack_depth,
            "image_age_category": image_age_category,
            "session": session,
        }

        tiles = batch_search_by_area(**kwargs)
        self.results = SelectResult.from_geojson(tiles)
        self.succeeded = True
        self.failed = False
        self.running = False
        self.finished = True
        self.status = "SUCCEEDED"
        self.state = "SUCCEEDED"

    def submit() -> None:
        pass

    def wait_for_success(self, interval: int = 5) -> None:
        pass

    not_implemented = {
        "select_id": "MetaSelects use batch search and do not have an ID",
        "from_id": "MetaSelects use batch search and do not have an ID",
        "get_select": "MetaSelects use batch search and do not have an ID",
        "usage": "For faster response time, MetaSelects do not calculate usage",
        "get_signed_link": "MetaSelects do not generate files, see .results()",
        "get_link_contents": "MetaSelects do not generate files, see .results()",
        "copy_link": "MetaSelects do not generate files, see .results()",
    }

    def __getattr__(self, val):
        if val in self.not_implemented:
            raise NotImplemented(self.not_implemented[val])

    def __repr__(self) -> str:
        return f"<ARD MetaSelect>"


###
# Cell Metadata
###


def search_by_cell_id(
    cell_id: str,
    acq_ids: List[str] = [],
    datetime: str = None,
    intersects=None,
    bbox: list = None,
    query: dict = {},
    stack_depth: int = None,
    image_age_category: List[str] = [],
    format: str = "geojson",
    session=None,
) -> dict:

    """Searches for tile metadata within a given ARD grid cell.

    Parameters
    ----------
    cell_id : str
        An ARD tile cell ID, like Z10-302310230201
    acq_ids : iterable of str
        An iterable of acquisition IDs to search for
    datetime : str
        Date or date range string
    intersects : Geometry-like objects, str (optional))
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north]
    query : dict
        Query dictionary
    stack_depth: int or None, optional
        Maximum number of tiles to return, default is to return all tiles
    image_age_category : iterable of str, optional
        One or
    session : session object, optional
        A user_session or application_session object
    format: str
        either 'geojson' (default), 'order', or 'stac'

    Returns
    -------
    dict
        output dictionary, see Notes

    Notes
    -----
    Default stack depth is unlimited, use `stack_depth` to set a limit on tiles returned. There is no scoring on parameters, so if results are limited by passing a stack depth, tiles are returned sorted newest to oldest. A `stack_depth` of 5 will return the five most recent tiles.

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading

    Formats:
        "geojson": GeoJSON FeatureCollection format, each tile is a Feature
        "order": A dictionary used by the order system: keys are acquisition IDs, values are lists of cell IDs
        "stac": STAC search API collection format"""

    url = ard_url("metadata", "cells", cell_id, f"acquisitions?format={format}")

    if intersects is not None:
        intersects = convert_to_shapely(intersects).wkt

    payload = {
        "ids": acq_ids,
        "datetime": datetime,
        "intersects": intersects,
        "bbox": bbox,
        "stack_depth": stack_depth,
        "query": query,
        "image_age_category": image_age_category,
    }
    payload = {k: v for k, v in payload.items() if v not in [None, [], ()]}

    session = session or get_user_session()
    return session.post(url, json=payload).json()


def search_by_area(
    acq_ids: List[str] = [],
    datetime: str = None,
    intersects=None,
    bbox: list = None,
    query: dict = {},
    stack_depth: int = None,
    image_age_category: List[str] = [],
    format: str = "geojson",
    session=None,
) -> dict:
    """Searches for tile metadata within a given area.

    Parameters
    ----------
    acq_ids : iterable of str
        An iterable of acquisition IDs to search for
    datetime : str
        Date or date range string
    intersects : Geometry-like objects, str
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north]
    query : dict
        Query dictionary
    stack_depth: int or None, optional
        Maximum number of tiles to return
    image_age_category : iterable of str, optional
        One or
    session : session object, optional
        A user_session or application_session object
    format: str
        either 'geojson' (default), 'order', or 'stac'

    Returns
    -------
    dict
        output dictionary, see Notes

    Notes
    -----
    Default stack depth is unlimited, use `stack_depth` to set a limit on tiles returned. There is no scoring on parameters, so if results are limited by passing a stack depth, tiles are returned sorted newest to oldest.

    A geometry is required (either `intersects` or `bbox`) if acquisition IDs are not passed. If IDs are passed without a geometry, it is assumed all tiles from those acquisitions should be returned.

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading

    Formats:
        "geojson": GeoJSON FeatureCollection format, each tile is a Feature
        "order": A dictionary used by the order system: keys are acquisition IDs, values are lists of cell IDs
        "stac": STAC search API collection format"""

    url = ard_url("metadata", "cells", f"acquisitions?format={format}")

    if intersects is not None:
        intersects = convert_to_shapely(intersects).wkt

    payload = {
        "ids": acq_ids,
        "datetime": datetime,
        "intersects": intersects,
        "bbox": bbox,
        "stack_depth": stack_depth,
        "query": query,
        "image_age_category": image_age_category,
    }
    payload = {k: v for k, v in payload.items() if v not in [None, [], ()]}

    session = session or get_user_session()
    return session.post(url, json=payload).json()


def batch_search_by_area(filters: list = [], session=None, **kwargs) -> dict:
    """A multithreaded batch search for tile metadata over a given area.

    - the search area is broken down to individual tiles
    - the tiles are searched asychronously in 8 parallel threads
    - custom filter functions can be run on each batch in the asychronous process
    - only returns the GeoJSON format

    Parameters
    ----------
    filters : list of functions
        A list of functions to use to filter tiles (see Notes)
    acq_ids : iterable of str
        An iterable of acquisition IDs to search for
    datetime : str
        Date or date range string
    intersects : Geometry-like objects
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north] if `intersects` is not provided
    query : dict
        Query dictionary
    stack_depth: int or None, optional
        Maximum number of tiles to return, default is all tiles
    image_age_category : iterable of str, optional
        One or more of "training", "standard", "fresh"
    session : session object, optional
        A user_session or application_session object


    Returns
    -------
    dict
        GeoJSON FeatureCollection format, each tile is a Feature dictionary.

    Notes
    -----
    Stack depth:
        Default stack depth is unlimited, use `stack_depth` to set a limit on tiles returned.
        Tiles are scored only on recency and you will receive the most recent matching tiles
        up to your stack depth.

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading

    Filter functions:
        These should take the GeoJSON Feature formatted tile dictionary as an input, and return True
        if the tile meets the filter requirements. For example:

            def match_worldview(tile):
                # check if the word "world" is in the platform name
                return "world" in tile["properties"]["platform"].lower()

        A tile must pass all filters to be returned (they are logically ANDed). Filters execute in order,
        so more expensive filters can be placed later in the list.

    """

    # get the AOI
    if "bbox" in kwargs:
        if kwargs["bbox"] is not None:
            aoi = box(*kwargs["bbox"])
        del kwargs["bbox"]
    if "intersects" in kwargs:
        if kwargs["intersects"] is not None:
            aoi = convert_to_shapely(kwargs["intersects"])
        del kwargs["intersects"]

    prep_aoi = prep(aoi)
    fetches = []
    for cell in covers(aoi):
        url = ard_url("metadata", "cells", cell.id, "acquisitions?format=geojson")
        payload = kwargs.copy()
        payload = {k: v for k, v in payload.items() if v not in [None, [], ()]}
        intersection = aoi.intersection(cell.geom_WGS84)
        # skip intersections that are touching but don't have an area
        if not intersection.is_empty:
            # interior cells that are covered don't need an AOI
            if not prep_aoi.contains_properly(intersection):
                # since the intersection can be less than the 1 sqkm limit, send the full geometry
                payload["intersects"] = aoi.wkt
            fetches.append((url, payload))

    def filter_hook(response, *args, **kwargs):
        tiles = response.json()["features"]
        if len(filters) > 0:
            filter_fn = lambda tile: all([fn(tile) for fn in filters])
            response.tiles = filter(filter_fn, tiles)
        else:
            response.tiles = tiles

    session = session or get_user_session()

    tiles = []

    # requests-futures lets you "add" hooks as a kwarg
    # but these clobber existing hooks
    # so we'll add the filter hook to the hooks and then reset them back
    old_hooks = deepcopy(session.hooks)
    session.hooks["response"].append(filter_hook)
    try:
        with FuturesSession(session=session) as future_session:
            futures = [future_session.post(url, json=payload) for url, payload in fetches]
            for future in as_completed(futures):
                try:
                    tiles.extend(future.result().tiles)
                # tiles that are overlapped by an area too small to order are skipped
                except GeometryException as g:
                    raise g
    except Exception as e:
        raise e
    finally:
        session.hooks = old_hooks
    return {"type": "FeatureCollection", "features": tiles}


def get_acquisition(
    acq_id: str,
    format: str = "geojson",
    session=None,
) -> dict:
    """Get all ARD metadata for an acquisition

    Parameters
    ----------
    acq_id : str
        Acquisition ID for the tile
    session : session object, optional
        A user_session or application_session object
    format: str
        either 'geojson' (default), 'order', or 'stac'

    Returns
    -------
    dict
        output dictionary, see Notes

    Notes
    -----
    Formats:
        "geojson": GeoJSON FeatureCollection format, each tile is a Feature
        "order": A dictionary used by the order system: keys are acquisition IDs, values are lists of cell IDs
        "stac": STAC search API collection format"""

    return search_by_area(acq_ids=[acq_id], format=format, session=session)


def get_tile(
    cell_id: str,
    acq_id: str,
    session=None,
) -> dict:

    """Get a specific ARD tile's metadata

    Parameters
    ----------

    cell_id : str
        ARD grid cell ID for the tile
    acq_id : str
        Acquisition ID for the tile
    session : session object, optional
        A user_session or application_session object

    Returns
    -------
    dict
        GeoJSON Feature format of the ARD tile
    """

    url = ard_url("metadata", "cells", cell_id, "acquisitions", acq_id)

    session = session or get_user_session()
    return session.get(url).json()


####
# Summary Service
####


def summarize_cell(cell_id: str, session=None) -> dict:
    """Get a specific ARD cell's summary metadata

    Parameters
    ----------
    cell_id : str
        ARD grid cell ID
    session : session object, optional
        A user_session or application_session object

    Returns
    -------
    dict
        Summary metadata

    Raises
    ------
    MissingSummary
        The cell is not in an area indexed by the summary service.
        Any area can be added to the summary service using `register_summary_area()`
    """

    url = ard_url("metadata", "cells", cell_id, "summary")

    session = session or get_user_session()
    try:
        response = session.get(url)
    except MissingARDResource:
        msg = (
            f"Cell {cell_id} is not being indexed by the metadata summary service. \n"
            + "Use register_summary_coverage(geometry) to index this area and try again."
        )
        raise MissingSummary(msg)
    return response.json()


def summarize_area(intersects=None, bbox: list = None, strict: bool = False, session=None) -> dict:
    """Get a ARD cell summary metadata over an area

    Parameters
    ----------
    intersects : Geometry-like objects
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north] if `intersects` is not provided
    strict : bool, default is False
        Use strict mode. In strict mode missing summary tiles will raise MissingSummary exceptions
    session : session object, optional
        A user_session or application_session object

    Returns
    -------
    dict
        Summary metadata

    Raises
    ------
    MissingSummary
        If `strict=True`, some portion of the input area is not indexed by the summary service.
        Any area can be added to the index service by using `register_summary_area()`

    Notes
    -----
    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading
    """
    url = ard_url("metadata", "cells", "summary")

    payload = {}
    if bbox is not None:
        payload["bbox"] = bbox
    if intersects is not None:
        intersects = convert_to_shapely(intersects).wkt
        payload["intersects"] = intersects

    session = session or get_user_session()
    results = session.post(url, json=payload).json()
    if strict:
        intersects = intersects or box(*bbox)
        coverage = list(covers(intersects))
        if len(coverage) != len(results["features"]):
            raise MissingSummary(
                "The area requested does not have full coverage of summary data. \n"
                + "Register your area using register_summary_area() and try again."
            )
    return results


def register_summary_area(intersects=None, bbox: List = None, session=None) -> dict:
    """Register an area to be indexed by the summary service

    Parameters
    ----------
    intersects : Geometry-like objects
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north] if `intersects` is not provided
    session : session object, optional
        A user_session or application_session object

    Notes
    -----
    The summary service was initally indexing major metropolitan areas. You can add any additional areas you require.
    Summary areas do not need to be managed or deleted. Resubmitting an area or overlapping a previous one does not cause
    duplication.

    Intersects inputs:
        Geometry objects: Shapely shapes, GeoJSON-like dicts, GeoJSON and wkt strings,
            objects supporting __geo_interface__ like maxar_ard_grid.Cell
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading
    """
    url = ard_url("metadata", "cells", "summary", "register")

    payload = {}
    if bbox is not None:
        payload["bbox"] = bbox
    if intersects is not None:
        intersects = convert_to_shapely(intersects).wkt
        payload["intersects"] = intersects

    session = session or get_user_session()
    return session.post(url, json=payload).json()
